use std::ffi::OsString;
use std::fs::File;
use std::io::{Read, Write};
use std::os::{
    fd::FromRawFd,
    unix::process::{parent_id, CommandExt},
};
use std::process::{exit, Command};
use std::time::Duration;

use anyhow::{bail, Context};
use clap::Parser;
use nix::sys::wait::waitpid;
use nix::unistd::{fork, pipe, setsid, ForkResult, Pid};
use tokio::time::timeout;
use zbus::fdo::NameOwnerChangedStream;
use zbus::{export::futures_util::stream::StreamExt, fdo::DBusProxy, names::BusName, Connection};

#[derive(Parser)]
#[clap(author)]
struct Arguments {
    #[arg(short, long)]
    notify_fd: i32,

    #[arg(short, long)]
    bus_name: String,

    #[arg(short, long, default_value = "10")]
    timeout_secs: u64,

    prog: OsString,
    args: Vec<OsString>,
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> anyhow::Result<()> {
    let cli = Arguments::parse();

    let (reader, writer) = pipe().context("creating pipe")?;
    // SAFETY: The passed fd is a valid and open file descriptor.
    let reader = unsafe { File::from_raw_fd(reader) };
    let writer = unsafe { File::from_raw_fd(writer) };

    let mut cmd = Command::new(&cli.prog);
    cmd.args(&cli.args);

    // SAFETY: The passed fd is a valid and open file descriptor.
    let notify_fd = unsafe { File::from_raw_fd(cli.notify_fd) };

    let timeout = Duration::from_secs(cli.timeout_secs);

    // SAFETY: This is only unsafe for multithreaded programs. This context is single-thread.
    match unsafe { fork()? } {
        ForkResult::Parent { child } => parent(cmd, child, reader),
        ForkResult::Child => child(cli.bus_name, timeout, notify_fd, writer).await,
    }
}

fn parent(mut cmd: Command, child: Pid, mut r: impl Read) -> anyhow::Result<()> {
    // Reap the child ourselves since the service might not deal with that.
    waitpid(child, None)?;

    // Wait for child to signal readiness.
    let mut buffer = [0; 1];
    r.read(&mut buffer)?;
    if buffer != [b'\n'; 1] {
        bail!("child process did not signal readiness");
    }

    // Returns only if 'exec' calls fails:
    anyhow::bail!(cmd.exec());
}

async fn child(
    bus_name: String,
    timeout_duration: Duration,
    notify_fd: File,
    mut w: impl Write,
) -> anyhow::Result<()> {
    let parent_pid = parent_id();
    setsid()?;

    // SAFETY: This is only unsafe for multithreaded programs. This context is single-thread.
    if unsafe { fork()? }.is_parent() {
        exit(0);
    };

    let conn = Connection::session()
        .await
        .context("connecting to user d-bus")?;
    let proxy = DBusProxy::new(&conn).await.context("creating dbus proxy")?;
    let owner_changed = proxy
        .receive_name_owner_changed_with_args(&[(0, &bus_name)])
        .await?;

    // Indicate to the parent that the waiter is ready.
    w.write(&[b'\n'])?;
    drop(w); // Explicit close.

    let wait_future = wait_for_bus(proxy, owner_changed, parent_pid, notify_fd);
    timeout(timeout_duration, wait_future).await?
}

async fn wait_for_bus(
    proxy: DBusProxy<'_>,
    mut owner_changed: NameOwnerChangedStream<'_>,
    parent_pid: u32,
    mut notify_fd: impl Write,
) -> anyhow::Result<()> {
    while let Some(event) = owner_changed.next().await {
        if let Some(new_owner) = event.args()?.new_owner.take() {
            let new_owner = proxy
                .get_connection_unix_process_id(BusName::Unique(new_owner))
                .await
                .context("getting pid of new bus owner")?;

            if new_owner != parent_pid {
                bail!("another process took over the specified bus");
            }

            return notify_fd
                .write_all(b"READY=1\n")
                .context("writing to readiness fd");
        }
        // Else: no owner.
    }
    bail!("listening to signals stopped; did the d-bus die?");
}
